@@ -1,7 +1,7 @@
 #' Extract the headers from a response
 #'
 #' @param x A request object
-#' @seealso \code{\link{add_headers}()} to send additional headers in a
+#' @seealso [add_headers()] to send additional headers in a
 #'   request
 #' @export
 #' @examples
@@ -17,13 +17,13 @@ headers.response <- function(x) {
 #' Add additional headers to a request.
 #'
 #' Wikipedia provides a useful list of common http headers:
-#' \url{http://en.wikipedia.org/wiki/List_of_HTTP_header_fields}.
+#' <https://en.wikipedia.org/wiki/List_of_HTTP_header_fields>.
 #'
 #' @param ... named header values.  To stop an existing header from being
-#'   set, pass an empty string: \code{""}.
+#'   set, pass an empty string: `""`.
 #' @param .headers a named character vector
 #' @export
-#' @seealso \code{\link{accept}} and \code{\link{content_type}} for
+#' @seealso [accept()] and [content_type()] for
 #'   convenience functions for setting accept and content-type headers.
 #' @family config
 #' @examples
@@ -33,33 +33,28 @@ headers.response <- function(x) {
 #' GET("http://httpbin.org/headers")
 #'
 #' # Add arbitrary headers
-#' GET("http://httpbin.org/headers",
-#'  add_headers(version = version$version.string))
+#' GET(
+#'   "http://httpbin.org/headers",
+#'   add_headers(version = version$version.string)
+#' )
 #'
 #' # Override default headers with empty strings
 #' GET("http://httpbin.org/headers", add_headers(Accept = ""))
 add_headers <- function(..., .headers = character()) {
-  headers <- c(..., .headers)
-  if (length(headers) == 0) return()
-  stopifnot(is.character(headers))
-
-  # Keep last of duplicated headers
-  headers <- headers[!duplicated(names(headers), fromLast = TRUE)]
-
-  config(httpheader = headers)
+  request(headers = c(..., .headers))
 }
 
 
 #' Set content-type and accept headers.
 #'
-#' These are convenient wrappers aroud \code{\link{add_headers}}.
+#' These are convenient wrappers aroud [add_headers()].
 #'
-#' \code{accept_json}/\code{accept_xml} and
-#' \code{content_type_json}/\code{content_type_xml} are useful shortcuts to
+#' `accept_json`/`accept_xml` and
+#' `content_type_json`/`content_type_xml` are useful shortcuts to
 #' ask for json or xml responses or tell the server you are sending json/xml.
 #'
 #' @param type A mime type or a file extension. If a file extension (i.e. starts
-#'   with \code{.}) will guess the mime type using \code{\link[mime]{guess_type}}.
+#'   with `.`) will guess the mime type using [mime::guess_type()].
 #' @export
 #' @examples
 #' GET("http://httpbin.org/headers")
@@ -98,7 +93,6 @@ accept <- function(type) {
     type <- mime::guess_type(type, empty = NULL)
   }
   add_headers("Accept" = type)
-
 }
 
 #' @export
@@ -109,3 +103,45 @@ accept_json <- function() accept("application/json")
 #' @rdname content_type
 accept_xml <- function() accept("application/xml")
 
+# Parses a header lines as recieved from libcurl. Multiple responses
+# will be intermingled, each separated by an http status line.
+parse_http_headers <- function(raw) {
+  lines <- strsplit(rawToChar(raw), "\r?\n")[[1]]
+
+  new_response <- grepl("^HTTP", lines)
+  grps <- cumsum(new_response)
+
+  lapply(unname(split(lines, grps)), parse_single_header)
+}
+
+parse_single_header <- function(lines) {
+  status <- parse_http_status(lines[[1]])
+
+  # Parse headers into name-value pairs
+  header_lines <- lines[lines != ""][-1]
+  pos <- regexec("^([^:]*):\\s*(.*)$", header_lines)
+  pieces <- regmatches(header_lines, pos)
+
+  n <- vapply(pieces, length, integer(1))
+  if (any(n != 3)) {
+    bad <- header_lines[n != 3]
+    pieces <- pieces[n == 3]
+
+    warning("Failed to parse headers:\n", paste0(bad, "\n"), call. = FALSE)
+  }
+
+  names <- vapply(pieces, "[[", 2, FUN.VALUE = character(1))
+  values <- lapply(pieces, "[[", 3)
+  headers <- insensitive(stats::setNames(values, names))
+
+  list(status = status$status, version = status$version, headers = headers)
+}
+
+parse_http_status <- function(x) {
+  status <- as.list(strsplit(x, "\\s+")[[1]])
+  names(status) <- c("version", "status", "message")[seq_along(status)]
+  status$status <- as.integer(status$status)
+
+
+  status
+}
